@@ -110,7 +110,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 }]
                 pub async fn payment_webhook(
                     req_ctx: Arc<RequestContext<ServerContext>>,
-                    body: TypedBody<String>,
+                    body: UntypedBody,
                 ) -> Result<HttpResponseOk<()>, HttpError> {
                     let internal_ctx = req_ctx
                         .context()
@@ -126,23 +126,28 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                     let guard = req.lock().await;
                     let val = guard.deref();
                     let headers = val.headers().clone();
-                    let data = body.into_inner();
 
-                    let (correlation_id, args) = super::#pwh_name(headers, data).await.map_err(|e| {
+                    let maybe_callback_args: Option<CallbackArgs> = super::#pwh_name(headers, body).await.map_err(|e| {
                         let out: HttpError = e.into();
                         out
                     })?;
-                    ctx.start_transaction().await.unwrap();
-                    match callback_handler(&internal_ctx, &mut ctx, &correlation_id, args).await {
-                        Ok(_) => {
-                            ctx.commit_transaction().await.unwrap();
-                            Ok(HttpResponseOk(()))
-                        }
-                        Err(err) => {
-                            ctx.abort_transaction().await.unwrap();
-                            Err(err.into())
-                        }
+
+                    if let Some(callback_args) = maybe_callback_args {
+                        internal_ctx
+                            .payment_callbacks_topic
+                            .clone()
+                            .publish(callback_args)
+                            .await
+                            .map_err(|_| {
+                                let out: HttpError = new_application_error(
+                                    "UPSTREAM_ERROR",
+                                    "couldn't send message to payment callbacks topic",
+                                ).into();
+                                out
+                            })?;
                     }
+
+                    Ok(HttpResponseOk(()))
                 }
             };
 
@@ -183,11 +188,12 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
             use checkout_core::dropshot::RequestContext;
             use checkout_core::dropshot::Path;
             use checkout_core::dropshot::TypedBody;
+            use checkout_core::dropshot::UntypedBody;
             use checkout_core::dropshot::HttpError;
             use checkout_core::dropshot::HttpResponseOk;
             use checkout_core::http::StatusCode;
             use checkout_core::checkout::Checkout;
-            use checkout_core::payment::Payment;
+            use checkout_core::payment::{Payment, CallbackArgs};
             use checkout_core::payment::PaymentProcessor;
             use checkout_core::customer::Contact;
             use checkout_core::customer::Address;
@@ -213,7 +219,6 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 get_handler as get_checkout_handler,
             };
             use checkout_core::payment::handlers::{
-                process_handler,
                 step_handler,
                 callback_handler,
                 cancel_handler as cancel_payment_handler,
@@ -230,7 +235,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
 
             pub struct Config {
                 pub google_application_credentials: String,
-                pub payment_processing_topic: String,
+                pub payment_callbacks_topic: String,
                 pub order_events_topic: String,
                 pub bind_address: String,
             }
@@ -238,7 +243,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
             pub fn get_config() -> Config {
                 Config {
                     google_application_credentials: env::var("GOOGLE_APPLICATION_CREDENTIALS").unwrap(),
-                    payment_processing_topic: env::var("PAYMENT_PROCESSING_TOPIC").unwrap(),
+                    payment_callbacks_topic: env::var("PAYMENT_CALLBACKS_TOPIC").unwrap(),
                     order_events_topic: env::var("ORDER_EVENTS_TOPIC").unwrap(),
                     bind_address: env::var("BIND_ADDRESS").unwrap(),
                 }
@@ -284,17 +289,17 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 }
             }
 
-            pub fn start_payment_processing(
+            pub fn process_payment_callbacks(
                 internal_ctx: InternalContext,
                 context_provider: Arc<MasterContextProvider>,
             ) {
                 task::spawn(async move {
-                    let subscription = internal_ctx.payment_processing_subscription.clone();
+                    let subscription = internal_ctx.payment_callbacks_subscription.clone();
 
                     while subscription.client().is_running() {
                         println!("pulling payment processing messages");
 
-                        let pull = subscription.get_messages::<IdWrapper>();
+                        let pull = subscription.get_messages::<CallbackArgs<<super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs>>();
 
                         match timeout(Duration::from_secs(30), pull).await {
                             Ok(result) => match result {
@@ -302,18 +307,24 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                                     for (result, ack_id) in messages {
                                         match result {
                                             Ok(message) => {
-                                                println!("Received: {:?}", message);
+                                                println!("Received CallbackArgs message");
                                                 match context_provider.new_context().await {
                                                     Ok(mut ctx) => {
                                                         if !ctx.start_transaction().await.is_ok() {
                                                             continue;
                                                         }
-                                                        let processing_result = process_handler(&internal_ctx, &mut ctx, &message.0).await;
+                                                        let processing_result = callback_handler(&internal_ctx, &mut ctx, &message.correlation_id, message.process_args).await;
                                                         if !processing_result.is_ok() {
                                                             if !ctx.abort_transaction().await.is_ok() {
                                                                 // just log the failure to abort tx
-                                                                eprintln!("failed to commit tx");
+                                                                eprintln!("failed to abort tx");
                                                             }
+
+                                                            let processing_err = processing_result.err().unwrap();
+                                                            if &processing_err.code != "SERVER_ERROR" {
+                                                                subscription.acknowledge_messages(vec![ack_id]).await;
+                                                            }
+
                                                             continue;
                                                         }
                                                         if !ctx.commit_transaction().await.is_ok() {
@@ -336,7 +347,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 });
             }
 
-            pub fn start_order_reconciliation(
+            pub fn process_order_events(
                 internal_ctx: InternalContext,
                 context_provider: Arc<MasterContextProvider>,
             ) {
@@ -419,8 +430,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 body: TypedBody<CreateCheckoutParams>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>>, HttpError> {
                 let params = body.into_inner();
                 let mut ctx = req_ctx
@@ -451,8 +461,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>>, HttpError> {
                 let path_params = path.into_inner();
                 let mut ctx = req_ctx
@@ -484,8 +493,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 path: Path<IdParam>,
                 body: TypedBody<AddOrRemoveItemParams>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>>, HttpError> {
                 let path_params = path.into_inner();
                 let body_params = body.into_inner();
@@ -518,8 +526,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 path: Path<IdParam>,
                 body: TypedBody<AddOrRemoveItemParams>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>>, HttpError> {
                 let path_params = path.into_inner();
                 let body_params = body.into_inner();
@@ -551,8 +558,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>>, HttpError> {
                 let path_params = path.into_inner();
                 let mut ctx = req_ctx
@@ -585,8 +591,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 path: Path<IdParam>,
                 body: TypedBody<Contact>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>>, HttpError> {
                 let path_params = path.into_inner();
                 let contact = body.into_inner();
@@ -619,8 +624,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 path: Path<IdParam>,
                 body: TypedBody<Address>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>>, HttpError> {
                 let path_params = path.into_inner();
                 let shipping_address = body.into_inner();
@@ -653,8 +657,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 path: Path<IdParam>,
                 body: TypedBody<FulfillmentTypeSelection>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>>, HttpError> {
                 let path_params = path.into_inner();
                 let fulfillment_type = body.into_inner();
@@ -687,8 +690,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 path: Path<IdParam>,
                 body: TypedBody<<super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>>, HttpError> {
                 let path_params = path.into_inner();
                 let args = body.into_inner();
@@ -721,8 +723,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 path: Path<IdParam>,
                 body: TypedBody<<super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>>, HttpError> {
                 let path_params = path.into_inner();
                 let args = body.into_inner();
@@ -760,8 +761,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 path: Path<IdParam>,
                 body: TypedBody<<super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>>, HttpError> {
                 let path_params = path.into_inner();
                 let args = body.into_inner();
@@ -797,8 +797,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
             ) -> Result<HttpResponseOk<Checkout<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>>, HttpError> {
                 let path_params = path.into_inner();
                 let internal_ctx = req_ctx
@@ -835,8 +834,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
             ) -> Result<HttpResponseOk<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>, HttpError> {
                 let path_params = path.into_inner();
                 let mut ctx = req_ctx
@@ -868,8 +866,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 path: Path<IdParam>,
                 body: TypedBody<<super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs>,
             ) -> Result<HttpResponseOk<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>, HttpError> {
                 let path_params = path.into_inner();
                 let args = body.into_inner();
@@ -898,45 +895,6 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 }
             }
 
-            #[endpoint {
-                method = PATCH,
-                path = "/payment_callback/{correlation_id}",
-            }]
-            pub async fn callback_payment(
-                req_ctx: Arc<RequestContext<ServerContext>>,
-                path: Path<CorrelationIdParam>,
-                body: TypedBody<<super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs>,
-            ) -> Result<HttpResponseOk<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
-            >>, HttpError> {
-                let path_params = path.into_inner();
-                let args = body.into_inner();
-                let internal_ctx = req_ctx
-                    .context()
-                    .internal_context
-                    .clone();
-
-                let mut ctx = req_ctx
-                    .context()
-                    .context_provider
-                    .new_context()
-                    .await
-                    .unwrap();
-
-                ctx.start_transaction().await.unwrap();
-                match callback_handler(&internal_ctx, &mut ctx, &path_params.correlation_id, args).await {
-                    Ok(co) => {
-                        ctx.commit_transaction().await.unwrap();
-                        Ok(HttpResponseOk(co))
-                    }
-                    Err(err) => {
-                        ctx.abort_transaction().await.unwrap();
-                        Err(err.into())
-                    }
-                }
-            }
-
             #[endpoint {
                 method = PATCH,
                 path = "/payment/{id}/cancel",
@@ -945,8 +903,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 req_ctx: Arc<RequestContext<ServerContext>>,
                 path: Path<IdParam>,
             ) -> Result<HttpResponseOk<Payment<
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data,
-                <super::#context_type_name #ty_static_generics as PaymentProcessor>::ProcessArgs
+                <super::#context_type_name #ty_static_generics as PaymentProcessor>::Data
             >>, HttpError> {
                 let path_params = path.into_inner();
                 let internal_ctx = req_ctx
@@ -996,7 +953,6 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 api.register(checkout_server::release_checkout).unwrap();
                 api.register(checkout_server::get_payment).unwrap();
                 api.register(checkout_server::step_payment).unwrap();
-                api.register(checkout_server::callback_payment).unwrap();
                 api.register(checkout_server::cancel_payment).unwrap();
 
                 #payment_webhook_endpoint_reg
@@ -1015,7 +971,7 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
 
                 let api = Self::api_description();
                 let config = checkout_server::get_config();
-                let internal_context = checkout_server::InternalContext::new(config.google_application_credentials.clone(), config.payment_processing_topic.clone(), config.order_events_topic.clone()).await?;
+                let internal_context = checkout_server::InternalContext::new(config.google_application_credentials.clone(), config.payment_callbacks_topic.clone(), config.order_events_topic.clone()).await?;
 
                 let db_client = checkout_server::DBClient::new().await.map_err(|_| checkout_server::new_server_error("failed to create database client"))?;
                 let context_provider = checkout_server::Arc::new(checkout_server::MasterContextProvider::new(
@@ -1026,8 +982,8 @@ fn impl_checkout_server(ast: &DeriveInput) -> TokenStream {
                 let pubsub_client = internal_context.pubsub_client.clone();
                 pubsub_client.spawn_token_renew(checkout_server::Duration::from_secs(15 * 60));
 
-                checkout_server::start_payment_processing(internal_context.clone(), context_provider.clone());
-                checkout_server::start_order_reconciliation(internal_context.clone(), context_provider.clone());
+                checkout_server::process_payment_callbacks(internal_context.clone(), context_provider.clone());
+                checkout_server::process_order_events(internal_context.clone(), context_provider.clone());
 
                 checkout_server::HttpServerStarter::new(
                     &checkout_server::ConfigDropshot {
